package me.hammerle.snuviengine.api;

import java.nio.FloatBuffer;
import me.hammerle.snuviengine.util.Color;
import org.lwjgl.BufferUtils;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL30.*;

public class ColorRenderer
{
    private int vao;
    private int vbo;
    
    private float color;
    
    private FloatBuffer buffer = BufferUtils.createFloatBuffer(9);
    
    private boolean built = false;
    
    public ColorRenderer()
    {
        Shader.addTask(() -> 
        {
            vao = glGenVertexArrays();
            vbo = glGenBuffers();

            glBindVertexArray(vao);
            glBindBuffer(GL_ARRAY_BUFFER, vbo);

            glEnableVertexAttribArray(0);
            glVertexAttribPointer(0, 2, GL_FLOAT, false, 12, 0);

            glEnableVertexAttribArray(2);  
            glVertexAttribPointer(2, 4, GL_UNSIGNED_BYTE, true, 12, 8);
        });
    }
    
    public void setColor(float r, float g, float b, float a)
    {
        this.color = Float.intBitsToFloat(Color.get(r * 255.0f, g * 255.0f, b * 255.0f, a * 255.0f));
    }
    
    public void addTriangle(float x1, float y1, float x2, float y2, float x3, float y3)
    {
        if(buffer.position() + 9 > buffer.capacity())
        {
            FloatBuffer newBuffer = BufferUtils.createFloatBuffer(buffer.capacity() * 2);
            buffer.flip();
            newBuffer.put(buffer);
            buffer = newBuffer;
        }
        
        buffer.put(x1);
        buffer.put(y1);        
        buffer.put(color);
        
        buffer.put(x2);
        buffer.put(y2);        
        buffer.put(color);
        
        buffer.put(x3);
        buffer.put(y3);        
        buffer.put(color);
    }
    
    public void addRectangle(float minX, float minY, float maxX, float maxY)
    {
        addTriangle(minX, maxY, minX, minY, maxX, maxY);
        addTriangle(maxX, maxY, minX, minY, maxX, minY);
    }
    
    public void build()
    {
        if(!Shader.initDone || buffer.position() == 0)
        {
            throw new ShaderException("build called too early");
        }
        buffer.flip();
        glBindVertexArray(vao);
        glBindBuffer(GL_ARRAY_BUFFER, vbo);
        glBufferData(GL_ARRAY_BUFFER, buffer, GL_STATIC_DRAW);
        
        buffer.limit(buffer.capacity());
        built = true;
    }
    
    public void draw()
    {
        if(!built)
        {
            throw new ShaderException("build must be called before drawing");
        }
        glBindVertexArray(vao);
        glBindBuffer(GL_ARRAY_BUFFER, vbo);
        glDrawArrays(GL_TRIANGLES, 0, buffer.limit() / 3);
    }
    
    public void delete()
    {
        buffer = null;
        glDeleteVertexArrays(vao);
        glDeleteBuffers(vbo);
        vao = -1;
        vbo = -1;
        built = false;
    }
}
