package me.hammerle.snuviengine.api;

public class MatrixStack
{
    private final int capacity;
    private final float[][] stack;
    private int index = 0;
    
    protected MatrixStack(int capacity)
    {
        this.capacity = capacity;
        stack = new float[capacity][16];
        
        stack[index][0] = 1.0f;
        stack[index][5] = 1.0f;
        stack[index][10] = 1.0f;
        stack[index][15] = 1.0f;
    }
    
    protected float[] getData()
    {
        return stack[index];
    }
    
    protected void push()
    {
        index++;
        if(index >= capacity)
        {
            throw new ShaderException("matrix stack overflow, max capacity is " + capacity);
        }
        System.arraycopy(stack[index - 1], 0, stack[index], 0, 16);
    }
    
    protected void pop()
    {
        if(index <= 0)
        {
            throw new ShaderException("matrix stack underflow");
        }
        index--;
    }
    
    protected void scale(float sx, float sy) 
    {
        float[] data = stack[index];
        data[0] *= sx;
        data[1] *= sx;
        data[4] *= sy;
        data[5] *= sy;
    }
    
    protected void translate(float tx, float ty) 
    {
        float[] data = stack[index];
        data[12] += data[0] * tx + data[4] * ty;
        data[13] += data[1] * tx + data[5] * ty;
    }
    
    protected void translateTo(float tx, float ty) 
    {
        float[] data = stack[index];
        data[0] = 1.0f;
        data[1] = 0.0f;
        data[4] = 0.0f;
        data[5] = 1.0f; 
        data[12] = tx;
        data[13] = ty;
    }
    
    protected void rotate(float angle) 
    {
        float[] data = stack[index];
        
        angle = (float) Math.toRadians(angle);
        float cos = (float) Math.cos(angle);
        float sin = (float) Math.sin(angle);
        
        float a1 = data[0];
        float a2 = data[4];
        data[0] = a1 * cos - a2 * sin;
        data[4] = a1 * sin + a2 * cos;
        
        a1 = data[1];
        a2 = data[5];
        data[1] = a1 * cos - a2 * sin;
        data[5] = a1 * sin + a2 * cos;
    }
}
